from Products.CMFCore.utils import getToolByName


def enableExplicitMode():
    """ZCML startup hook to put the ruleset registry into explicit mode.
    This means we require people to declare ruleset types before using them.
    """
    from z3c.caching.registry import getGlobalRulesetRegistry

    registry = getGlobalRulesetRegistry()
    if registry is not None:
        registry.explicit = True


def importVarious(context):
    if not context.readDataFile("plone.app.caching.txt"):
        return

    site = context.getSite()

    error_log = getToolByName(site, "error_log")

    properties = error_log.getProperties()
    ignored = properties.get("ignored_exceptions", ())

    modified = False
    for exceptionName in ("Intercepted",):
        if exceptionName not in ignored:
            ignored += (exceptionName,)
            modified = True

    if modified:
        error_log.setProperties(
            properties.get("keep_entries", 10),
            properties.get("copy_to_zlog", True),
            ignored,
        )



from Acquisition import Explicit
from datetime import datetime
from dateutil.tz import tzlocal
from io import StringIO
from plone.registry.fieldfactory import persistentFieldAdapter
from plone.testing.zca import UNIT_TESTING
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.interfaces import IMembershipTool
from z3c.caching.interfaces import ILastModified
from zope.component import adapter
from zope.component import provideAdapter
from zope.component import provideUtility
from zope.interface import implementer
from zope.interface import Interface
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse

import time
import unittest


@implementer(IContentish)
class DummyContext(Explicit):
    def getLayout(self):
        return getattr(self, "_layout", None)


class DummyPublished:
    def __init__(self, parent=None):
        self.__parent__ = parent


class TestETags(unittest.TestCase):
    layer = UNIT_TESTING

    def setUp(self):
        provideAdapter(persistentFieldAdapter)

    # UserID

    def test_UserID_anonymous(self):
        from plone.app.caching.operations.etags import UserID

        @adapter(Interface, Interface)
        @implementer(IMembershipTool)
        class DummyMembershipTool:
            def __init__(self):
                pass

            def getAuthenticatedMember(self):
                return

        provideUtility(DummyMembershipTool())

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = UserID(published, request)

        self.assertIsNone(etag())

    def test_UserID_member(self):
        from plone.app.caching.operations.etags import UserID

        class DummyMember:
            def getId(self):
                return "bob"

        @adapter(Interface, Interface)
        @implementer(IMembershipTool)
        class DummyMembershipTool:
            def __init__(self):
                pass

            def getAuthenticatedMember(self):
                return DummyMember()

        provideUtility(DummyMembershipTool())

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = UserID(published, request)

        self.assertEqual("bob", etag())

    # Roles

    def test_Roles_anonymous(self):
        from plone.app.caching.operations.etags import Roles

        @adapter(Interface, Interface)
        @implementer(IMembershipTool)
        class DummyMembershipTool:
            def __init__(self):
                pass

            def getAuthenticatedMember(self):
                return

            def isAnonymousUser(self):
                return True

        provideUtility(DummyMembershipTool())

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = Roles(published, request)

        self.assertEqual("Anonymous", etag())

    def test_Roles_member(self):
        from plone.app.caching.operations.etags import Roles

        class DummyMember:
            def getRolesInContext(self, context):
                return ["Member", "Manager"]

        @adapter(Interface, Interface)
        @implementer(IMembershipTool)
        class DummyMembershipTool:
            def __init__(self):
                pass

            def isAnonymousUser(self):
                return False

            def getAuthenticatedMember(self):
                return DummyMember()

        provideUtility(DummyMembershipTool())

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = Roles(published, request)

        self.assertEqual("Manager;Member", etag())

    # Language

    def test_Language_no_header(self):
        from plone.app.caching.operations.etags import Language

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = Language(published, request)

        self.assertEqual("", etag())

    def test_Language_with_header(self):
        from plone.app.caching.operations.etags import Language

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        request.environ["HTTP_ACCEPT_LANGUAGE"] = "en"

        etag = Language(published, request)

        self.assertEqual("en", etag())

    # UserLanguage

    def test_UserLanguage(self):
        from plone.app.caching.operations.etags import UserLanguage

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        context = DummyContext()
        context.Language = lambda: "en"
        published = DummyPublished(context)

        etag = UserLanguage(published, request)

        self.assertEqual("en", etag())

    # LastModified

    def test_LastModified_no_adapter(self):
        from plone.app.caching.operations.etags import LastModified

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = LastModified(published, request)

        self.assertIsNone(etag())

    def test_LastModified_None(self):
        from plone.app.caching.operations.etags import LastModified

        @implementer(ILastModified)
        @adapter(DummyPublished)
        class DummyLastModified:
            def __init__(self, context):
                self.context = context

            def __call__(self):
                return

        provideAdapter(DummyLastModified)

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = LastModified(published, request)

        self.assertIsNone(etag())

    def test_LastModified(self):
        from plone.app.caching.operations.etags import LastModified

        mod = datetime(2010, 1, 2, 3, 4, 5, 6, tzlocal())
        utcStamp = time.mktime(mod.utctimetuple())

        @implementer(ILastModified)
        @adapter(DummyPublished)
        class DummyLastModified:
            def __init__(self, context):
                self.context = context

            def __call__(self):
                return mod

        provideAdapter(DummyLastModified)

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = LastModified(published, request)
        self.assertEqual(str(utcStamp), etag())

    # CatalogCounter

    def test_CatalogCounter(self):
        from plone.app.caching.operations.etags import CatalogCounter
        from Products.CMFCore.interfaces import ICatalogTool

        @implementer(ICatalogTool)
        class DummyCatalog:
            def getCounter(self):
                return 10

        provideUtility(DummyCatalog())

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = CatalogCounter(published, request)

        self.assertEqual("10", etag())

    # ObjectLocked

    def test_ObjectLocked_true(self):
        from plone.app.caching.operations.etags import ObjectLocked

        @implementer(Interface)
        @adapter(DummyContext, Interface)
        class DummyContextState:
            def __init__(self, context, request):
                pass

            def is_locked(self):
                return True

        provideAdapter(DummyContextState, name="plone_context_state")

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = ObjectLocked(published, request)

        self.assertEqual("1", etag())

    def test_ObjectLocked_false(self):
        from plone.app.caching.operations.etags import ObjectLocked

        @implementer(Interface)
        @adapter(DummyContext, Interface)
        class DummyContextState:
            def __init__(self, context, request):
                pass

            def is_locked(self):
                return False

        provideAdapter(DummyContextState, name="plone_context_state")

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = ObjectLocked(published, request)

        self.assertEqual("0", etag())

    # Skin

    def test_Skin_request_variable(self):
        from plone.app.caching.operations.etags import Skin

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        class DummyPortalSkins:
            def getRequestVarname(self):
                return "skin_name"

            def getDefaultSkin(self):
                return "defaultskin"

        published.__parent__.portal_skins = DummyPortalSkins()
        request["skin_name"] = "otherskin"

        etag = Skin(published, request)

        self.assertEqual("otherskin", etag())

    def test_Skin_default(self):
        from plone.app.caching.operations.etags import Skin

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        class DummyPortalSkins:
            def getRequestVarname(self):
                return "skin_name"

            def getDefaultSkin(self):
                return "defaultskin"

        published.__parent__.portal_skins = DummyPortalSkins()

        etag = Skin(published, request)

        self.assertEqual("defaultskin", etag())

    # layout

    def test_layout_none(self):
        from plone.app.caching.operations.etags import Layout

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        published = DummyPublished(DummyContext())

        etag = Layout(published, request)

        self.assertIsNone(etag())

    def test_layout_set(self):
        from plone.app.caching.operations.etags import Layout

        environ = {"SERVER_NAME": "example.com", "SERVER_PORT": "80"}
        response = HTTPResponse()
        request = HTTPRequest(StringIO(), environ, response)
        context = DummyContext()
        context._layout = "hello_view"
        published = DummyPublished(context)

        etag = Layout(published, request)

        self.assertEqual("hello_view", etag())



import os
import time


class TZEnvContext:
    def __init__(self, tzval):
        self.tzval = tzval
        self._old_tz = None

    def get_current_tz(self):
        return os.environ.get("TZ", None)

    def set_current_tz(self, tzval):
        if tzval is None and "TZ" in os.environ:
            del os.environ["TZ"]
        else:
            os.environ["TZ"] = tzval

        time.tzset()

    def __enter__(self):
        self._old_tz = self.get_current_tz()
        self.set_current_tz(self.tzval)

    def __exit__(self, type, value, traceback):
        if self._old_tz is not None:
            self.set_current_tz(self._old_tz)

        self._old_tz = None



from plone.app.caching.interfaces import IRAMCached
from plone.app.caching.operations.utils import storeResponseInRAMCache
from plone.transformchain.interfaces import ITransform
from zope.component import adapter
from zope.interface import implementer
from zope.interface import Interface


GLOBAL_KEY = "plone.app.caching.operations.ramcache"


@implementer(ITransform)
@adapter(Interface, Interface)
class Store:
    """Transform chain element which actually saves the page in RAM.

    This is registered for the ``IRAMCached`` request marker, which is set by
    the ``cacheInRAM()`` helper method. Thus, the transform is only used if
    the caching operation requested it.
    """

    order = 90000

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transformUnicode(self, result, encoding):
        if self.responseIsSuccess() and IRAMCached.providedBy(self.request):
            storeResponseInRAMCache(
                self.request, self.request.response, result.encode(encoding)
            )

    def transformBytes(self, result, encoding):
        if self.responseIsSuccess() and IRAMCached.providedBy(self.request):
            storeResponseInRAMCache(self.request, self.request.response, result)

    def transformIterable(self, result, encoding):
        if self.responseIsSuccess() and IRAMCached.providedBy(self.request):
            result = b"".join(result)
            storeResponseInRAMCache(self.request, self.request.response, result)
            # ITransform contract allows to return an "encoded string" aka bytes
            return result

    def responseIsSuccess(self):
        return self.request.response.getStatus() == 200



from Acquisition import aq_parent
from plone.app.caching.utils import getObjectDefaultView
from plone.app.caching.utils import isPurged
from plone.cachepurging.interfaces import IPurgePathRewriter
from plone.dexterity.content import get_assignable
from plone.dexterity.interfaces import IDexteritySchema
from plone.dexterity.schema import SCHEMA_CACHE
from plone.memoize.instance import memoize
from plone.namedfile.interfaces import INamedBlobFileField
from plone.namedfile.interfaces import INamedImageField
from plone.registry.interfaces import IRegistry
from Products.CMFCore.interfaces import IContentish
from Products.CMFCore.interfaces import IDiscussionResponse
from Products.CMFCore.interfaces import IDynamicType
from Products.CMFCore.utils import getToolByName
from z3c.caching.interfaces import IPurgePaths
from z3c.caching.purge import Purge
from zope.component import adapter
from zope.component import getAdapters
from zope.component import getUtility
from zope.event import notify
from zope.globalrequest import getRequest
from zope.interface import implementer
from zope.lifecycleevent.interfaces import IObjectAddedEvent
from zope.lifecycleevent.interfaces import IObjectModifiedEvent
from zope.lifecycleevent.interfaces import IObjectMovedEvent
from zope.lifecycleevent.interfaces import IObjectRemovedEvent
from zope.schema import getFieldsInOrder

import pkg_resources


try:
    pkg_resources.get_distribution("plone.restapi")
    HAS_RESTAPI = True
except pkg_resources.DistributionNotFound:
    HAS_RESTAPI = False

CONTENT_PATHS_POSTFIXES = [
    "/view",
]
if HAS_RESTAPI:
    CONTENT_PATHS_POSTFIXES += [
        "/@comments",
    ]


def _append_paths(paths, prefix=""):
    for postfix in CONTENT_PATHS_POSTFIXES:
        paths.append(prefix + postfix)


@implementer(IPurgePaths)
@adapter(IDynamicType)
class ContentPurgePaths:
    """Paths to purge for content items

    Includes:

    * ${object_path}/ (e.g. for folders)
    * ${object_path}/view
    * ${object_path}/${object_default_view}
    * a bunch of restapi endpoints

    If the object is the default view of its parent, also purge:

    * ${parent_path}
    * ${parent_path}/
    """

    def __init__(self, context):
        self.context = context

    def getRelativePaths(self):
        prefix = "/" + self.context.virtual_url_path()
        paths = [prefix + "/", prefix + "/view"]

        defaultView = getObjectDefaultView(self.context)
        if defaultView:
            path = prefix + "/" + defaultView
            if path not in paths:  # it could be
                paths.append(path)

        parent = aq_parent(self.context)
        if parent is None:
            return paths

        parentDefaultView = getObjectDefaultView(parent)
        if parentDefaultView == self.context.getId():
            parentPrefix = "/" + parent.virtual_url_path()
            paths.append(parentPrefix)
            if parentPrefix == "/":
                # special handling for site root since parentPrefix
                # does not make sense in that case.
                # Additionally, empty site roots were not getting
                # purge paths
                # /VirtualHostBase/http/site.com:80/site1/VirtualHostRoot/_vh_site1/
                # was getting generated but not
                # /VirtualHostBase/http/site.com:80/site1/VirtualHostRoot/_vh_site1
                # which would translate to http://site.come/ getting
                # invalidated but not http://site.come
                paths.append("")
                _append_paths(paths)
            else:
                paths.append(parentPrefix + "/")
                _append_paths(paths, prefix=parentPrefix)

        return paths

    def getAbsolutePaths(self):
        return []


@implementer(IPurgePaths)
@adapter(IDiscussionResponse)
class DiscussionItemPurgePaths:
    """Paths to purge for Discussion Item.

    Looks up paths for the ultimate parent.
    """

    def __init__(self, context):
        self.context = context

    def getRelativePaths(self):
        root = self._getRoot()
        if root is None:
            return

        request = getRequest()
        if request is None:
            return

        rewriter = IPurgePathRewriter(request, None)
        for _name, pathProvider in getAdapters((root,), IPurgePaths):
            # add relative paths, which are rewritten
            relativePaths = pathProvider.getRelativePaths()
            if relativePaths:
                for relativePath in relativePaths:
                    if rewriter is None:
                        yield relativePath
                        continue
                    rewrittenPaths = rewriter(relativePath) or []
                    yield from rewrittenPaths

    def getAbsolutePaths(self):
        root = self._getRoot()
        if root is None:
            return

        request = getRequest()
        if request is None:
            return

        for _name, pathProvider in getAdapters((root,), IPurgePaths):
            # add absolute paths, which are not
            absolutePaths = pathProvider.getAbsolutePaths()
            if absolutePaths:
                yield from absolutePaths

    @memoize
    def _getRoot(self):
        plone_utils = getToolByName(self.context, "plone_utils", None)
        if plone_utils is None:
            return

        thread = plone_utils.getDiscussionThread(self.context)
        if not thread:
            return

        return thread[0]


@implementer(IPurgePaths)
@adapter(IDexteritySchema)
class ScalesPurgePaths:
    """Paths to purge for Dexterity object fields"""

    def __init__(self, context):
        self.context = context

    def getScales(self):
        if getattr(self, "_sizes", None) is None:
            reg_list = getUtility(IRegistry)["plone.allowed_sizes"]
            self._sizes = [i.split(" ", 1)[0] for i in reg_list]
        return self._sizes

    def getRelativePaths(self):
        prefix = "/" + self.context.virtual_url_path()

        def schemas():
            yield SCHEMA_CACHE.get(self.context.portal_type)
            for behavior_registration in get_assignable(
                self.context
            ).enumerateBehaviors():
                yield behavior_registration.interface

        for schema in schemas():
            for field_name, field in getFieldsInOrder(schema):
                is_image = INamedImageField.providedBy(field)
                is_file = INamedBlobFileField.providedBy(field)
                if not (is_image or is_file):
                    continue
                value = field.get(self.context)
                if not value:
                    continue
                filename = value.filename
                if is_image:
                    yield f"{prefix}/@@images/{field_name}"
                    for size in self.getScales():
                        yield f"{prefix}/images/{field_name}/{size}"
                        yield f"{prefix}/@@images/{field_name}/{size}"
                if is_file:
                    yield f"{prefix}/view/++widget++form.widgets.{field_name}/@@download/{filename}"
                yield f"{prefix}/download/{field_name}"
                yield f"{prefix}/download/{field_name}/{filename}"
                yield f"{prefix}/@@download/{field_name}"
                yield f"{prefix}/@@download/{field_name}/{filename}"

    def getAbsolutePaths(self):
        return []


# Event redispatch for content items - we check the list of content items
# instead of the marker interface


@adapter(IContentish, IObjectModifiedEvent)
def purgeOnModified(object, event):
    if isPurged(object):
        notify(Purge(object))


@adapter(IContentish, IObjectMovedEvent)
def purgeOnMovedOrRemoved(object, event):
    request = getRequest()
    confirmed_delete = (
        request is not None
        and getattr(request, "URL", None)
        and "delete_confirmation" in request.URL
        and request.REQUEST_METHOD == "POST"
        and "form.submitted" in request.form
    )
    if not confirmed_delete and IObjectRemovedEvent.providedBy(event):
        # ignore extra delete events
        return
    # Don't purge when added
    if IObjectAddedEvent.providedBy(event):
        return
    if isPurged(object) and "portal_factory" not in request.URL:
        notify(Purge(object))
    parent = object.getParentNode()
    if parent:
        notify(Purge(parent))



from io import BytesIO
from plone.app.caching.interfaces import IPloneCacheSettings
from plone.app.caching.testing import PLONE_APP_CACHING_FUNCTIONAL_TESTING
from plone.app.caching.tests.test_utils import normalize_etag
from plone.app.caching.tests.test_utils import stable_now
from plone.app.caching.tests.test_utils import test_image
from plone.app.testing import applyProfile
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import TEST_USER_PASSWORD
from plone.app.textfield.value import RichTextValue
from plone.cachepurging.interfaces import ICachePurgingSettings
from plone.cachepurging.interfaces import IPurger
from plone.caching.interfaces import ICacheSettings
from plone.registry.interfaces import IRegistry
from plone.testing.z2 import Browser
from Products.CMFCore.FSFile import FSFile
from Products.CMFCore.utils import getToolByName
from zope.component import getUtility
from zope.globalrequest import setRequest

import datetime
import dateutil.parser
import dateutil.tz
import os
import transaction
import unittest


class TestProfileWithCaching(unittest.TestCase):
    """This test aims to exercise the caching operations expected from the
    `with-caching-proxy` profile.

    Several of the operations are just duplicates of the ones for the
    `without-caching-proxy` profile but we still want to redo them in this
    context to ensure the profile has set the correct settings.
    """

    layer = PLONE_APP_CACHING_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]

        test_css = FSFile(
            "test.css", os.path.join(os.path.dirname(__file__), "test.css")
        )
        self.portal.portal_skins.custom._setOb("test.css", test_css)

        setRequest(self.portal.REQUEST)

        applyProfile(self.portal, "plone.app.caching:with-caching-proxy")

        self.registry = getUtility(IRegistry)

        self.cacheSettings = self.registry.forInterface(ICacheSettings)
        self.cachePurgingSettings = self.registry.forInterface(ICachePurgingSettings)
        self.ploneCacheSettings = self.registry.forInterface(IPloneCacheSettings)

        self.cacheSettings.enabled = True

        self.purger = getUtility(IPurger)
        self.purger.reset()

    def tearDown(self):
        setRequest(None)

    def test_composite_viewsxx(self):
        # This is a clone of the same test for 'without-caching-proxy'
        # Can we just call that test from this context?

        catalog = self.portal["portal_catalog"]
        skins_tool = self.portal["portal_skins"]

        # Add folder content
        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.portal.invokeFactory("Folder", "f1")
        self.portal["f1"].title = "one"
        self.portal["f1"].description = "Folder one description"
        self.portal["f1"].reindexObject()

        # Add page content
        self.portal["f1"].invokeFactory("Document", "d1")
        self.portal["f1"]["d1"].title = "Document one"
        self.portal["f1"]["d1"].description = "Document one description"
        testText = "Testing... body one"
        self.portal["f1"]["d1"].text = RichTextValue(
            testText,
            "text/plain",
            "text/html",
        )
        self.portal["f1"]["d1"].reindexObject()

        # Publish the folder and page
        self.portal.portal_workflow.doActionFor(self.portal["f1"], "publish")
        self.portal.portal_workflow.doActionFor(self.portal["f1"]["d1"], "publish")

        # Should we set up the etag components?
        # - set member?  No
        # - reset catalog counter?  Maybe
        # - set server language?
        # - turn on gzip?
        # - set skin?  Maybe
        # - leave status unlocked
        #
        transaction.commit()

        # Request the authenticated folder
        now = stable_now()
        browser = Browser(self.app)
        browser.addHeader(
            "Authorization",
            f"Basic {TEST_USER_NAME}:{TEST_USER_PASSWORD}",
        )
        browser.open(self.portal["f1"].absolute_url())
        self.assertEqual("plone.content.folderView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        # This should use cacheInBrowser
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        self.assertEqual(
            f'"|test_user_1_|{catalog.getCounter()}|en|{skins_tool.default_skin}|0|0|"',
            normalize_etag(browser.headers["ETag"]),
        )
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Set the copy/cut cookie and then request the folder view again
        browser.cookies.create("__cp", "xxx")
        browser.open(self.portal["f1"].absolute_url())
        # The response should be the same as before except for the etag
        self.assertEqual("plone.content.folderView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        self.assertEqual(
            f'"|test_user_1_|{catalog.getCounter()}|en|{skins_tool.default_skin}|0|1|"',
            normalize_etag(browser.headers["ETag"]),
        )

        # Request the authenticated page
        now = stable_now()
        browser = Browser(self.app)
        browser.addHeader(
            "Authorization",
            f"Basic {TEST_USER_NAME}:{TEST_USER_PASSWORD}",
        )
        browser.open(self.portal["f1"]["d1"].absolute_url())
        self.assertIn(testText, browser.contents)
        self.assertEqual("plone.content.itemView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        # This should use cacheInBrowser
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        self.assertEqual(
            f'"|test_user_1_|{catalog.getCounter()}|en|{skins_tool.default_skin}|0|"',
            normalize_etag(browser.headers["ETag"]),
        )
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request the authenticated page again -- to test RAM cache.
        browser = Browser(self.app)
        browser.addHeader(
            "Authorization",
            f"Basic {TEST_USER_NAME}:{TEST_USER_PASSWORD}",
        )
        browser.open(self.portal["f1"]["d1"].absolute_url())
        self.assertEqual("plone.content.itemView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        # Authenticated should NOT be RAM cached
        self.assertIsNone(browser.headers.get("X-RAMCache"))

        # Request the authenticated page again -- with an INM header to test
        # 304
        etag = browser.headers["ETag"]
        browser = Browser(self.app)
        browser.raiseHttpErrors = False  # we really do want to see the 304
        browser.addHeader(
            "Authorization",
            f"Basic {TEST_USER_NAME}:{TEST_USER_PASSWORD}",
        )
        browser.addHeader("If-None-Match", etag)
        browser.open(self.portal["f1"]["d1"].absolute_url())
        # This should be a 304 response
        self.assertEqual("304 Not Modified", browser.headers["Status"])
        # We MUST have an empty byte, not text, otherwise it is an indication that we
        # get a possibly wrong Content-Type in a 304 status.
        # See https://github.com/zopefoundation/Zope/issues/1089.
        self.assertEqual(b"", browser.contents)
        self.assertFalse(browser.headers["Content-Type"])

        # Request the anonymous folder
        now = stable_now()
        browser = Browser(self.app)
        browser.handleErrors = False
        browser.open(self.portal["f1"].absolute_url())
        self.assertEqual("plone.content.folderView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        # This should use cacheInBrowser
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        self.assertEqual(
            f'"||{catalog.getCounter()}|en|{skins_tool.default_skin}|0|0|"',
            normalize_etag(browser.headers["ETag"]),
        )
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request the anonymous page
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal["f1"]["d1"].absolute_url())
        self.assertEqual("plone.content.itemView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        self.assertIn(testText, browser.contents)
        # This should use cacheInBrowser
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        self.assertEqual(
            f'"||{catalog.getCounter()}|en|{skins_tool.default_skin}|0|"',
            normalize_etag(browser.headers["ETag"]),
        )
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request the anonymous page again -- to test RAM cache.
        # Anonymous should be RAM cached
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal["f1"]["d1"].absolute_url())
        self.assertEqual("plone.content.itemView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        # This should come from RAM cache
        self.assertEqual(
            "plone.app.caching.operations.ramcache", browser.headers["X-RAMCache"]
        )
        self.assertIn(testText, browser.contents)
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        self.assertEqual(
            f'"||{catalog.getCounter()}|en|{skins_tool.default_skin}|0|"',
            normalize_etag(browser.headers["ETag"]),
        )
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request the anonymous page again -- with an INM header to test 304.
        etag = browser.headers["ETag"]
        browser = Browser(self.app)
        browser.raiseHttpErrors = False
        browser.addHeader("If-None-Match", etag)
        browser.open(self.portal["f1"]["d1"].absolute_url())
        self.assertEqual("plone.content.itemView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        # This should be a 304 response
        self.assertEqual("304 Not Modified", browser.headers["Status"])
        self.assertEqual(b"", browser.contents)
        self.assertFalse(browser.headers["Content-Type"])

        # Edit the page to update the etag
        testText2 = "Testing... body two"
        self.portal["f1"]["d1"].text = RichTextValue(
            testText2,
            "text/plain",
            "text/html",
        )
        self.portal["f1"]["d1"].reindexObject()

        transaction.commit()

        # Request the anonymous page again -- to test expiration of 304 and
        # RAM.
        etag = browser.headers["ETag"]
        browser = Browser(self.app)
        browser.addHeader("If-None-Match", etag)
        browser.open(self.portal["f1"]["d1"].absolute_url())
        self.assertEqual("plone.content.itemView", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.weakCaching", browser.headers["X-Cache-Operation"]
        )
        # The etag has changed so we should get a fresh page.
        self.assertIsNone(browser.headers.get("X-RAMCache"))
        self.assertEqual("200 OK", browser.headers["Status"].upper())

    def test_content_feeds(self):
        catalog = self.portal["portal_catalog"]
        skins_tool = self.portal["portal_skins"]

        # Enable syndication
        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.syndication = getToolByName(self.portal, "portal_syndication")
        self.syndication.editProperties(isAllowed=True)
        self.syndication.enableSyndication(self.portal)

        transaction.commit()

        # Request the rss feed
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal.absolute_url() + "/RSS")
        self.assertEqual("plone.content.feed", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # This should use cacheInProxy
        self.assertEqual(
            "max-age=0, s-maxage=86400, must-revalidate",
            browser.headers["Cache-Control"],
        )
        tag = f'"||{catalog.getCounter()}|en|{skins_tool.default_skin}"'
        self.assertEqual(tag, browser.headers["ETag"])
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request the rss feed again -- to test RAM cache
        now = stable_now()
        rssText = browser.contents
        browser = Browser(self.app)
        browser.open(self.portal.absolute_url() + "/RSS")
        self.assertEqual("plone.content.feed", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # This should come from the RAM cache
        self.assertEqual(
            "plone.app.caching.operations.ramcache", browser.headers["X-RAMCache"]
        )
        self.assertEqual(rssText, browser.contents)
        self.assertEqual(
            "max-age=0, s-maxage=86400, must-revalidate",
            browser.headers["Cache-Control"],
        )
        tag = f'"||{catalog.getCounter()}|en|{skins_tool.default_skin}"'
        self.assertEqual(tag, browser.headers["ETag"])
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request the rss feed again -- with an INM header to test 304.
        etag = browser.headers["ETag"]
        browser = Browser(self.app)
        browser.raiseHttpErrors = False
        browser.addHeader("If-None-Match", etag)
        browser.open(self.portal.absolute_url() + "/RSS")
        self.assertEqual("plone.content.feed", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # This should be a 304 response
        self.assertEqual("304 Not Modified", browser.headers["Status"])
        self.assertEqual(b"", browser.contents)
        self.assertFalse(browser.headers["Content-Type"])

        # Request the authenticated rss feed
        now = stable_now()
        browser = Browser(self.app)
        browser.addHeader(
            "Authorization",
            f"Basic {TEST_USER_NAME}:{TEST_USER_PASSWORD}",
        )
        browser.open(self.portal.absolute_url() + "/RSS")
        self.assertEqual("plone.content.feed", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # This should use cacheInBrowser
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        tag = f'"|test_user_1_|{catalog.getCounter()}|en|{skins_tool.default_skin}"'
        self.assertEqual(tag, browser.headers["ETag"])
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request the authenticated rss feed again -- to test RAM cache
        browser = Browser(self.app)
        browser.addHeader(
            "Authorization",
            f"Basic {TEST_USER_NAME}:{TEST_USER_PASSWORD}",
        )
        browser.open(self.portal.absolute_url() + "/RSS")
        self.assertEqual("plone.content.feed", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # Authenticated should NOT be RAM cached
        self.assertIsNone(browser.headers.get("X-RAMCache"))

    def test_content_files(self):
        # Add folder content
        setRoles(self.portal, TEST_USER_ID, ("Manager",))
        self.portal.invokeFactory("Folder", "f1")
        self.portal["f1"].title = "Folder one"
        self.portal["f1"].description = "Folder one description"
        self.portal["f1"].reindexObject()

        # Add content image
        self.portal["f1"].invokeFactory("Image", "i1")
        self.portal["f1"]["i1"].title = "Image one"
        self.portal["f1"]["i1"].description = "Image one description"
        self.portal["f1"]["i1"].image = test_image()
        self.portal["f1"]["i1"].reindexObject()

        transaction.commit()

        # Request the image with Manager role
        now = stable_now()
        browser = Browser(self.app)
        browser.addHeader(
            "Authorization",
            f"Basic {SITE_OWNER_NAME}:{SITE_OWNER_PASSWORD}",
        )
        browser.open(self.portal["f1"]["i1"].absolute_url())
        self.assertEqual("plone.content.file", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # Folder not published yet so image should not be cached in proxy
        # so this should use cacheInBrowser
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get("Last-Modified"))
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request an image scale with Manager role
        now = stable_now()
        browser = Browser(self.app)
        browser.handleErrors = False
        browser.addHeader(
            "Authorization",
            f"Basic {SITE_OWNER_NAME}:{SITE_OWNER_PASSWORD}",
        )
        browser.open(self.portal["f1"]["i1"].absolute_url() + "/@@images/image/preview")
        self.assertEqual("plone.content.file", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # Folder not published yet so image scale should not be cached in proxy
        # so this should use cacheInBrowser
        self.assertEqual(
            "max-age=0, must-revalidate, private", browser.headers["Cache-Control"]
        )
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get("Last-Modified"))
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Publish the folder
        self.portal.portal_workflow.doActionFor(self.portal["f1"], "publish")

        transaction.commit()

        # Request the image
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal["f1"]["i1"].absolute_url())
        self.assertEqual("plone.content.file", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # Now visible to anonymous so this should use cacheInProxy
        self.assertEqual(
            "max-age=0, s-maxage=86400, must-revalidate",
            browser.headers["Cache-Control"],
        )
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get("Last-Modified"))
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

        # Request the image again -- with an IMS header to test 304
        lastmodified = browser.headers["Last-Modified"]
        browser = Browser(self.app)
        browser.raiseHttpErrors = False
        browser.addHeader("If-Modified-Since", lastmodified)
        browser.open(self.portal["f1"]["i1"].absolute_url())
        self.assertEqual("plone.content.file", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # This should be a 304 response
        self.assertEqual("304 Not Modified", browser.headers["Status"])
        self.assertEqual(b"", browser.contents)
        self.assertFalse(browser.headers["Content-Type"])

        # Request an image scale
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal["f1"]["i1"].absolute_url() + "/@@images/image/preview")
        self.assertEqual("plone.content.file", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.moderateCaching", browser.headers["X-Cache-Operation"]
        )
        # Now visible to anonymous so this should use cacheInProxy
        self.assertEqual(
            "max-age=0, s-maxage=86400, must-revalidate",
            browser.headers["Cache-Control"],
        )
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get("Last-Modified"))
        self.assertGreater(now, dateutil.parser.parse(browser.headers["Expires"]))

    def test_resources(self):
        # This is a clone of the same test for 'without-caching-proxy'
        # Can we just call that test from this context?

        transaction.commit()

        # Request a skin image
        now = stable_now()
        browser = Browser(self.app)
        browser.open(self.portal.absolute_url() + "/rss.png")
        self.assertEqual("plone.resource", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.strongCaching", browser.headers["X-Cache-Operation"]
        )
        # This should use cacheInBrowserAndProxy
        self.assertEqual(
            "max-age=86400, proxy-revalidate, public", browser.headers["Cache-Control"]
        )
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get("Last-Modified"))
        timedelta = dateutil.parser.parse(browser.headers["Expires"]) - now
        self.assertGreater(timedelta, datetime.timedelta(seconds=86390))

        # Request the skin image again -- with an IMS header to test 304
        lastmodified = browser.headers["Last-Modified"]
        browser = Browser(self.app)
        browser.raiseHttpErrors = False
        browser.addHeader("If-Modified-Since", lastmodified)
        browser.open(self.portal.absolute_url() + "/rss.png")
        self.assertEqual("plone.resource", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.strongCaching", browser.headers["X-Cache-Operation"]
        )
        # This should be a 304 response
        self.assertEqual("304 Not Modified", browser.headers["Status"])
        self.assertEqual(b"", browser.contents)
        self.assertFalse(browser.headers["Content-Type"])

        # Request a large datafile (over 64K) to test files that use
        # the "response.write()" function to initiate a streamed response.
        # This is of type OFS.Image.File but it should also apply to
        # large OFS.Image.Image, large non-blog ATImages/ATFiles, and
        # large Resource Registry cooked files, which all use the same
        # method to initiate a streamed response.
        s = b"a" * (1 << 16) * 3
        self.portal.manage_addFile(
            "bigfile", file=BytesIO(s), content_type="application/octet-stream"
        )

        transaction.commit()

        browser = Browser(self.app)
        browser.open(self.portal["bigfile"].absolute_url())
        self.assertEqual("plone.resource", browser.headers["X-Cache-Rule"])
        self.assertEqual(
            "plone.app.caching.strongCaching", browser.headers["X-Cache-Operation"]
        )
        # This should use cacheInBrowserAndProxy
        self.assertEqual(
            "max-age=86400, proxy-revalidate, public", browser.headers["Cache-Control"]
        )
        # remove this when the next line works
        self.assertIsNotNone(browser.headers.get("Last-Modified"))
        timedelta = dateutil.parser.parse(browser.headers["Expires"]) - now
        self.assertGreater(timedelta, datetime.timedelta(seconds=86390))

    def test_stable_resources(self):
        # This is a clone of the same test for 'without-caching-proxy'
        # Can we just call that test from this context?
        # (yes, this is not really testing anything. It's a placeholder)
        #
        # We don't actually have any non-RR stable resources yet
        # What is the best way to test this?
        # Maybe not important since the RR test exercises the same code?
        pass



from pathlib import Path
from setuptools import find_packages
from setuptools import setup


version = "3.0.4"

long_description = f"""
{Path("README.md").read_text()}\n
{Path("CHANGES.md").read_text()}\n
"""

setup(
    name="plone.app.caching",
    version=version,
    description="Plone UI and default rules for plone.caching/z3c.caching",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: Core",
        "Framework :: Zope :: 5",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="plone caching",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://github.com/plone/plone.app.caching",
    license="GPL version 2",
    packages=find_packages(),
    namespace_packages=["plone", "plone.app"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "AccessControl",
        "Acquisition",
        "Products.CMFCore",
        "Products.CMFDynamicViewFTI",
        "Products.CMFPlone",
        "Products.GenericSetup",
        "Products.statusmessages",
        "setuptools",
        "python-dateutil",
        "plone.app.registry",
        "plone.base",
        "plone.caching",
        "plone.dexterity",
        "plone.cachepurging",
        "plone.memoize",
        "plone.namedfile",
        "plone.protect",
        "plone.registry",
        "plone.transformchain",
        "plone.z3cform",
        "z3c.caching",
        "z3c.form",
        "z3c.zcmlhook",
        "zope.annotation",
        "zope.browserresource",
        "zope.component",
        "zope.event",
        "zope.globalrequest",
        "zope.i18nmessageid",
        "zope.interface",
        "zope.lifecycleevent",
        "zope.pagetemplate",
        "zope.publisher",
        "zope.ramcache",
        "zope.schema",
    ],
    extras_require={
        "test": [
            "persistent",
            "plone.app.contenttypes[test]",
            "plone.app.testing",
            "plone.app.textfield",
            "plone.behavior",
            "plone.testing",
            "pytz",
            "transaction",
        ]
    },
    entry_points="""
    [z3c.autoinclude.plugin]
    target = plone
    """,
)



from Acquisition import aq_base
from Acquisition import aq_inner
from plone.app.caching.interfaces import IETagValue
from plone.app.caching.operations.utils import getContext
from plone.app.caching.operations.utils import getLastModifiedAnnotation
from plone.base.utils import safe_hasattr
from Products.CMFCore.interfaces import ICatalogTool
from Products.CMFCore.interfaces import IMembershipTool
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.resources.utils import get_override_directory
from Products.CMFPlone.resources.utils import PRODUCTION_RESOURCE_DIRECTORY
from zope.component import adapter
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.interface import Interface

import random
import time


@implementer(IETagValue)
@adapter(Interface, Interface)
class UserID:
    """The ``userid`` etag component, returning the current user's id"""

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        tool = queryUtility(IMembershipTool)
        if tool is None:
            return

        member = tool.getAuthenticatedMember()
        if member is None:
            return

        return member.getId()


@implementer(IETagValue)
@adapter(Interface, Interface)
class Roles:
    """The ``roles`` etag component, returning the current user's roles,
    separated by semicolons
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        tool = queryUtility(IMembershipTool)
        if tool is None:
            return

        if bool(tool.isAnonymousUser()):
            return "Anonymous"

        member = tool.getAuthenticatedMember()
        if member is None:
            return

        return ";".join(sorted(member.getRolesInContext(getContext(self.published))))


@implementer(IETagValue)
@adapter(Interface, Interface)
class Language:
    """The ``language`` etag component, returning the value of the
    HTTP_ACCEPT_LANGUAGE request key.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        return self.request.get("HTTP_ACCEPT_LANGUAGE", "")


@implementer(IETagValue)
@adapter(Interface, Interface)
class UserLanguage:
    """The ``userLanguage`` etag component, returning the user's preferred
    language
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        language = self.request.get("LANGUAGE", None)
        if language:
            return language
        context = getContext(self.published)
        language = aq_inner(context).Language()
        if language:
            return language
        portal_state = queryMultiAdapter(
            (context, self.request), name="plone_portal_state"
        )
        if portal_state is None:
            return
        return portal_state.default_language()


@implementer(IETagValue)
@adapter(Interface, Interface)
class LastModified:
    """The ``lastModified`` etag component, returning the last modification
    timestamp
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        lastModified = getLastModifiedAnnotation(self.published, self.request)
        if lastModified is None:
            return
        return str(time.mktime(lastModified.utctimetuple()))


@implementer(IETagValue)
@adapter(Interface, Interface)
class CatalogCounter:
    """The ``catalogCounter`` etag component, returning a counter which is
    incremented each time the catalog is updated.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        catalog = queryUtility(ICatalogTool)
        if catalog is None:
            return
        return str(catalog.getCounter())


@implementer(IETagValue)
@adapter(Interface, Interface)
class ObjectLocked:
    """The ``locked`` etag component, returning 1 or 0 depending on whether
    the object is locked.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        context_state = queryMultiAdapter(
            (context, self.request), name="plone_context_state"
        )
        if context_state is None:
            return
        return "1" if context_state.is_locked() else "0"


@implementer(IETagValue)
@adapter(Interface, Interface)
class Skin:
    """The ``skin`` etag component, returning the current skin name."""

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)

        portal_skins = getToolByName(context, "portal_skins", None)
        if portal_skins is None:
            return

        requestVariable = portal_skins.getRequestVarname()
        if requestVariable in self.request:
            return self.request[requestVariable]

        return portal_skins.getDefaultSkin()


@implementer(IETagValue)
@adapter(Interface, Interface)
class AnonymousOrRandom:
    """The ``anonymousOrRandom`` etag component. This is normally added
    implicitly by the ``anonOnly`` setting. It will return for anonymous
    users, but a random number for logged-in ones. The idea is to force a
    re-fetch of a page every time for logged-in users.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        tool = queryUtility(IMembershipTool)
        if tool is None:
            return
        if bool(tool.isAnonymousUser()):
            return
        return f"{time.time()}{random.randint(0, 1000)}"


@implementer(IETagValue)
@adapter(Interface, Interface)
class CopyCookie:
    """The ``copy`` etag component, returning 1 or 0 depending on whether
    the '__cp' cookie is set.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        return "1" if self.request.get("__cp") else "0"


@implementer(IETagValue)
@adapter(Interface, Interface)
class ResourceRegistries:
    """The ``resourceRegistries`` etag component, returning a timestamp.

    This is the last modified timestamp from the Plone 5+ Resource Registries.
    This is useful for avoiding requests for expired resources from cached pages.
    """

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        container = get_override_directory(context)
        if PRODUCTION_RESOURCE_DIRECTORY not in container:
            return ""
        production_folder = container[PRODUCTION_RESOURCE_DIRECTORY]
        filename = "timestamp.txt"
        if filename not in production_folder:
            return ""
        timestamp = production_folder.readFile(filename)
        if not timestamp:
            return ""
        # timestamp is in bytes, and we must return a string.
        # On Python 2 this is the same, but not on Python 3.
        if not isinstance(timestamp, str):
            timestamp = timestamp.decode("utf-8")
        return timestamp


@implementer(IETagValue)
@adapter(Interface, Interface)
class Layout:
    """The 'layout' etag component, returning they layout of a content item."""

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def __call__(self):
        context = getContext(self.published)
        if not safe_hasattr(aq_base(context), "getLayout"):
            return
        return context.getLayout()
